from __future__ import annotations

from typing import TYPE_CHECKING
from typing import List as typingList
from typing import Type

from aasm.intermediate.action import SendMessageAction
from aasm.intermediate.behaviour import MessageReceivedBehaviour
from aasm.utils.validation import is_float, is_int

if TYPE_CHECKING:
    from parsing.state import State


class ArgumentType:
    ...


class Mutable(ArgumentType):
    ...


class Immutable(ArgumentType):
    ...


class Declared(ArgumentType):
    ...


class Float(ArgumentType):
    ...


class Integer(ArgumentType):
    ...


class Enum(ArgumentType):
    ...


class EnumValue(ArgumentType):
    ...


class List(ArgumentType):
    ...


class ConnectionList(ArgumentType):
    ...


class MessageList(ArgumentType):
    ...


class AgentParam(ArgumentType):
    ...


class Message(ArgumentType):
    ...


class MessageType(ArgumentType):
    ...


class ReceivedMessage(ArgumentType):
    ...


class ReceivedMessageParam(ArgumentType):
    ...


class SendMessage(ArgumentType):
    ...


class SendMessageParam(ArgumentType):
    ...
    

class Connection(ArgumentType):
    ...


class Argument:
    """Doesn't panic. Use in the action context."""
    
    def __init__(self, state: State, expr: str):
        self.expr: str = expr
        self.types: typingList[ArgumentType] = []
        self.type_in_op: ArgumentType | None = None
        self.set_types(state)
   
    def set_types(self, state: State) -> None:        
        self.check_agent_params(state)
        self.check_action_variables(state)
        self.check_received_message_params(state)
        self.check_send_message_params(state)
        self.check_numerical_values()
    
    def check_agent_params(self, state: State) -> None:
        if self.expr in state.last_agent.RESERVED_FLOAT_PARAMS:
            self.types.append(self.compose(Float, AgentParam, Immutable))
            
        elif self.expr in state.last_agent.RESERVED_CONNECTION_LIST_PARAMS:
            self.types.append(self.compose(List, ConnectionList, AgentParam, Mutable))
            
        elif self.expr in state.last_agent.float_param_names:
            self.types.append(self.compose(Float, AgentParam, Mutable))
            
        elif self.expr in state.last_agent.enums:
            self.types.append(self.compose(Enum, AgentParam, Mutable))
            
        elif self.expr in state.last_agent.message_lists:
            self.types.append(self.compose(List, MessageList, AgentParam, Mutable))
            
        elif self.expr in state.last_agent.connection_lists:
            self.types.append(self.compose(List, ConnectionList, AgentParam, Mutable))
            
        for enum_param in state.last_agent.enums.values():
            for enum_value in enum_param.enum_values:
                if self.expr == enum_value.value:
                    self.types.append(self.compose(EnumValue, Immutable, from_enum=enum_param.name))
        
    def check_action_variables(self, state: State) -> None:
        if state.last_action.is_declaration_in_scope(self.expr):
            self.types.append(self.compose(Float, Declared, Mutable))
    
    def check_numerical_values(self) -> None:
        if is_float(self.expr):
            self.types.append(self.compose(Float, Immutable))

        if is_int(self.expr):
            self.types.append(self.compose(Integer, Immutable))
        
    def check_received_message_params(self, state: State) -> None:
        if isinstance(state.last_behaviour, MessageReceivedBehaviour):
            if self.expr.lower().startswith('rcv.'):
                prop = self.expr.split('.')[1]
                
                if prop in state.last_behaviour.received_message.RESERVED_CONNECTION_PARAMS:
                    self.types.append(self.compose(Connection, ReceivedMessageParam, Immutable))
                    
                elif prop in state.last_behaviour.received_message.RESERVED_TYPE_PARAMS:
                    self.types.append(self.compose(MessageType, ReceivedMessageParam, Immutable))
                
                elif prop in state.last_behaviour.received_message.float_params:
                    self.types.append(self.compose(Float, ReceivedMessageParam, Immutable))
                    
            elif self.expr.lower() == 'rcv':
                self.types.append(self.compose(Message, ReceivedMessage, Immutable))
                
    def check_send_message_params(self, state: State) -> None:
        if isinstance(state.last_action, SendMessageAction):
            if self.expr.lower().startswith('send.'):
                prop = self.expr.split('.')[1]
                    
                if prop in state.last_action.send_message.RESERVED_TYPE_PARAMS:
                    self.types.append(self.compose(MessageType, SendMessageParam, Immutable))
                
                elif prop in state.last_action.send_message.float_params:
                    self.types.append(self.compose(Float, SendMessageParam, Mutable))
                    
            elif self.expr.lower() == 'send':
                self.types.append(self.compose(Message, SendMessage, Mutable))
    
    def compose(self, *classes: Type[ArgumentType], **args: str) -> ArgumentType:
        name = "_".join([klass.__name__ for klass in classes])
        if args:
            name += '|' +  "|".join([f'{key}={value}' for key, value in args.items()])
        return type(name, classes, args)()

    def has_arg(self, argument_type: ArgumentType, key: str, value: str) -> bool:
        try:
            if getattr(argument_type, key) == value:
                return True
        except AttributeError:
            ...
        return False

    def has_type(self, *classes: Type[ArgumentType], **args: str) -> bool:
        for type_ in self.types:
            if all([isinstance(type_, klass) for klass in classes]) and all([self.has_arg(type_, key, value) for key, value in args.items()]):
                return True
        return False
                    
    def set_op_type(self, *classes: Type[ArgumentType], **args: str) -> None:
        for type_ in self.types:
            if all([isinstance(type_, klass) for klass in classes]) and all([self.has_arg(type_, key, value) for key, value in args.items()]):
                self.type_in_op = type_

    def declaration_context(self, rhs: Argument) -> bool:
        if rhs.has_type(Float):
            self.type_in_op = self.compose(Float, Declared, Mutable)
            rhs.set_op_type(Float)
            return True
        
        return False
    
    def unordered_comparaison_context(self, rhs: Argument) -> bool:
        if self.has_type(Float) and rhs.has_type(Float):
            self.set_op_type(Float)
            rhs.set_op_type(Float)
            
        elif self.has_type(Enum) and rhs.has_type(EnumValue, from_enum=self.expr):
            self.set_op_type(Enum)
            rhs.set_op_type(EnumValue, from_enum=self.expr)
            
        else:
            return False
        
        return True

    def ordered_comparaison_context(self, rhs: Argument) -> bool:
        if self.has_type(Float) and rhs.has_type(Float):
            self.set_op_type(Float)
            rhs.set_op_type(Float)
            return True
        
        return False

    def math_context(self, rhs: Argument) -> bool:
        if self.has_type(Float, Mutable) and rhs.has_type(Float):
            self.set_op_type(Float, Mutable)
            rhs.set_op_type(Float)
            return True
        
        return False

    def list_modification_context(self, rhs: Argument) -> bool:
        if self.has_type(ConnectionList, Mutable) and rhs.has_type(Connection):
            self.set_op_type(ConnectionList, Mutable)
            rhs.set_op_type(Connection)
            
        elif self.has_type(MessageList, Mutable) and rhs.has_type(Message):
            self.set_op_type(MessageList, Mutable)
            rhs.set_op_type(Message)
            
        else:
            return False
        
        return True
    
    def list_n_removal_context(self, rhs: Argument) -> bool:
        if self.has_type(List, Mutable) and rhs.has_type(Float):
            self.set_op_type(List, Mutable)
            rhs.set_op_type(Float)
            return True
        
        return False
    
    def assignment_context(self, rhs: Argument) -> bool:
        if self.has_type(Enum, Mutable) and rhs.has_type(EnumValue, from_enum=self.expr):
            self.set_op_type(Enum, Mutable)
            rhs.set_op_type(EnumValue, from_enum=self.expr)
            
        elif self.has_type(Float, Mutable) and rhs.has_type(Float):
            self.set_op_type(Float, Mutable)
            rhs.set_op_type(Float)
            
        elif self.has_type(Message, Mutable) and rhs.has_type(MessageList):
            self.set_op_type(Message, Mutable)
            rhs.set_op_type(MessageList)
            
        elif self.has_type(SendMessageParam, Mutable) and rhs.has_type(Float):
            self.set_op_type(SendMessageParam, Mutable)
            rhs.set_op_type(Float)
            
        else:
            return False
        
        return True
    
    def list_subset_context(self, from_list: Argument, num: Argument) -> bool:
        if self.has_type(ConnectionList, Mutable) and from_list.has_type(ConnectionList) and num.has_type(Float):
            self.set_op_type(ConnectionList, Mutable)
            from_list.set_op_type(ConnectionList)
            num.set_op_type(Float)
            return True
        
        return False
    
    def list_inclusion_context(self, rhs: Argument) -> bool:
        if self.has_type(ConnectionList) and rhs.has_type(Connection):
            self.set_op_type(ConnectionList)
            rhs.set_op_type(Connection)
        
        elif self.has_type(MessageList) and rhs.has_type(Message):
            self.set_op_type(MessageList)
            rhs.set_op_type(Message)

        else:
            return False
        
        return True
    
    def list_clear_context(self) -> bool:
        if self.has_type(List, Mutable):
            self.set_op_type(List, Mutable)
            return True
        
        return False
    
    def list_length_context(self, rhs: Argument) -> bool:
        if self.has_type(Float, Mutable) and rhs.has_type(List):
            self.set_op_type(Float, Mutable)
            rhs.set_op_type(List)
            return True
        
        return False
    
    def send_context(self) -> bool:
        if self.has_type(ConnectionList):
            self.set_op_type(ConnectionList)
            
        elif self.has_type(Connection):
            self.set_op_type(Connection)
            
        else:
            return False
        
        return True

    def random_number_generation_context(self, *args: Argument) -> bool:
        if self.has_type(Float, Mutable) and all([arg.has_type(Float) for arg in args]):
            self.set_op_type(Float, Mutable)
            for arg in args:
                arg.set_op_type(Float)
            return True

        return False

    def round_number_context(self) -> bool:
        if self.has_type(Float, Mutable):
            self.set_op_type(Float, Mutable)
            return True

        return False

    def explain(self) -> str:
        types = f'{self.expr}: [ '
        for argument_type in self.types:
            types += type(argument_type).__name__ + ', '
        types = types.rstrip().rsplit(',', 1)[0]
        types += ' ]'
        return types

    def print(self) -> None:
        print(f'Argument {self.expr}')
        print(f'Type in op: {type(self.type_in_op).__name__}')
        for argument_type in self.types:
            type(argument_type).__name__



from __future__ import annotations

from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
    from aasm.intermediate.declaration import Declaration
    from aasm.intermediate.instruction import Instruction


class Block:    
    def __init__(self, names_declared_in_parent: List[str]):
        self.statements: List[Declaration | Instruction | Block] = []
        self._declared_names: List[str] = list(names_declared_in_parent)
        
    @property
    def declarations_in_scope(self) -> List[str]:
        return self._declared_names
        
    def add_declaration(self, declaration: Declaration) -> None:
        self._declared_names.append(declaration.name)
        self.statements.append(declaration)
    
    def add_statement(self, statement: Instruction | Block) -> None:
        self.statements.append(statement)
        
    def print(self) -> None:
        print(f'Block')
        print(f'Names in scope')
        print(self._declared_names)
        for instruction in self.statements:
            instruction.print()
        print('(EndBlock)')



import copy
import datetime
import random
import httpx
import numpy
import orjson
import spade
import sys


class average_user(spade.agent.Agent):
    def __init__(self, jid, password, backup_url = None, backup_period = 60, backup_delay = 0, logger = None, **kwargs):
        super().__init__(jid, password, verify_security=False)
        if logger: logger.debug(f'[{jid}] Received parameters: jid: {jid}, password: {password}, backup_url: {backup_url}, backup_period: {backup_period}, backup_delay: {backup_delay}, kwargs: {kwargs}')
        self.logger = logger
        self.backup_url = backup_url
        self.backup_period = backup_period
        self.backup_delay = backup_delay
        self.connections = kwargs.get("connections", [])
        self.msgRCount = self.limit_number(kwargs.get("msgRCount", 0))
        self.msgSCount = self.limit_number(kwargs.get("msgSCount", 0))
        self.friends = kwargs.get("friends", [])
        if self.logger: self.logger.debug(f'[{self.jid}] Class dict after initialization: {self.__dict__}')
    
    @property
    def connCount(self):
        return self.limit_number(len(self.connections))
    
    def limit_number(self, value):
        return float(max(sys.float_info.min, min(value, sys.float_info.max)))
    
    def get_json_from_spade_message(self, msg):
        return orjson.loads(msg.body)
    
    def get_spade_message(self, receiver_jid, body):
        msg = spade.message.Message(to=receiver_jid)
        body["sender"] = str(self.jid)
        msg.metadata["type"] = body["type"]
        msg.metadata["performative"] = body["performative"]
        msg.body = str(orjson.dumps(body), encoding="utf-8")
        return msg
    
    def setup(self):
        if self.backup_url:
            BackupBehaviour_template = spade.template.Template()
            BackupBehaviour_template.set_metadata("reserved", "no_message_match")
            self.add_behaviour(self.BackupBehaviour(start_at=datetime.datetime.now() + datetime.timedelta(seconds=self.backup_delay), period=self.backup_period), BackupBehaviour_template)
        facebook_activity_template = spade.template.Template()
        facebook_activity_template.set_metadata("reserved", "no_message_match")
        self.add_behaviour(self.facebook_activity(period=30), facebook_activity_template)
        if self.logger: self.logger.debug(f'[{self.jid}] Class dict after setup: {self.__dict__}')
    
    class BackupBehaviour(spade.behaviour.PeriodicBehaviour):
        def __init__(self, start_at, period):
            super().__init__(start_at=start_at, period=period)
            self.http_client = httpx.AsyncClient(timeout=period)
        
        async def run(self):
            data = {
                "jid": str(self.agent.jid),
                "type": "average_user",
                "floats": {
                    "msgRCount": self.agent.msgRCount,
                    "msgSCount": self.agent.msgSCount,
                    "connCount": self.agent.connCount,
                },
                "enums": {
                },
                "connections": {
                    "connections": self.agent.connections,
                    "friends": self.agent.friends,
                },
                "messages": {
                }
            }
            if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Sending backup data: {data}')
            try:
                await self.http_client.post(self.agent.backup_url, headers={"Content-Type": "application/json"}, data=orjson.dumps(data))
            except Exception as e:
                if self.agent.logger: self.agent.logger.error(f'[{self.agent.jid}] Backup error type: {e.__class__}, additional info: {e}')
    
    class facebook_activity(spade.behaviour.PeriodicBehaviour):
        async def post_photos(self):
            if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Run action post_photos')
            send = { "type": "facebook_post", "performative": "query", "photos": 0.0, }
            num_photos = self.agent.limit_number(0)
            num_photos = self.agent.limit_number(random.uniform(self.agent.limit_number(21), self.agent.limit_number(37)))
            num_photos = self.agent.limit_number(round(self.agent.limit_number(num_photos)))
            send["photos"] = self.agent.limit_number(num_photos)
            if self.agent.logger: self.agent.logger.debug(f'[{self.agent.jid}] Send message {send} to {self.agent.friends}')
            for receiver in self.agent.friends:
                await self.send(self.agent.get_spade_message(receiver, send))
                self.agent.msgSCount = self.agent.limit_number(self.agent.msgSCount + 1)
        
        async def run(self):
            await self.post_photos()
    

import random
import uuid
import numpy


def generate_graph_structure(domain):
    _num_average_user = round(100 / 100 * 150)
    num_agents = _num_average_user
    random_id = str(uuid.uuid4())[:5]
    jids = [f"{i}_{random_id}@{domain}" for i in range(num_agents)]
    agents = []
    next_agent_idx = 0
    for _ in range(_num_average_user):
        num_connections = int(numpy.random.exponential(1 / 0.1))
        num_connections = max(min(num_connections, len(jids) - 1), 0)
        jid = jids[next_agent_idx]
        agents.append({
            "jid": jid,
            "type": "average_user",
            "connections": random.sample([other_jid for other_jid in jids if other_jid != jid], num_connections),
        })
        next_agent_idx += 1
    return agents



from __future__ import annotations

from typing import TYPE_CHECKING, List

from aasm.intermediate.action import ModifySelfAction, SendMessageAction
from aasm.utils.validation import is_valid_name, print_invalid_names

if TYPE_CHECKING:
    from aasm.parsing.state import State


def op_ACTION(state: State, name: str, category: str, args: List[str]) -> None:
    state.require(
        state.in_behaviour, 
        'Actions must be definied inside behaviours.', 
        'Try defining new behaviours using BEHAV.'
    )
    state.require(
        not state.in_action, 
        'Cannot define an action in another action.', 
        'Try finishing current action using EACTION.'
    )
    state.require(not state.last_behaviour.action_exists(name), f'Action {name} already exists in current behaviour.')
    state.require(
        is_valid_name(name), 
        f'{name} is not a correct name.', 
        f'Names can only contain alphanumeric characters, underscores and cannot be: {print_invalid_names()}.'
    )
    
    match category, args:
        case 'modify_self', [ ]:
            state.last_behaviour.add_action(ModifySelfAction(name))
            
        case 'send_msg', [ msg_type, msg_performative ]:
            state.require(
                state.message_exists(msg_type, msg_performative), 
                f'Message {msg_type}/{msg_performative} does not exist.', 
                'Try defining new messages using MESSAGE.'
            )

            state.last_behaviour.add_action(SendMessageAction(name, state.get_message_instance(msg_type, msg_performative)))
            
        case _:
            state.panic(f'Incorrect operation: ACTION {category} {args}')
    
    state.in_action = True


def op_EACTION(state: State) -> None:            
    state.require(state.in_action, 'Not inside any action.', 'Try defining new actions using ACTION.')
    state.require(
        state.last_action.nested_blocks_count == 0, 
        'There are unclosed blocks.', 
        'Try closing them using EBLOCK.'
    )
    
    state.in_action = False



"""Agents Assembly translator"""

__version__ = "0.0.30"

from aasm.generating.code import Code
from aasm.generating.python_spade import get_spade_code
from aasm.utils.exception import PanicException
